from dataclasses import dataclass, field
import re


def clean_string(raw: str) -> str:
    """
    MQTT Discovery protocol only allows [a-zA-Z0-9_-]
    """
    result = re.sub(r"[^A-Za-z0-9_-]", "-", raw)
    return result.strip("-")


@dataclass
class Device:
    service: object
    name: str
    model: str | None = None
    manufacturer: str | None = None
    sw_version: str | None = None
    hw_version: str | None = None
    identifiers: list[str] | str | None = None
    connections: list[tuple] | None = None
    suggested_area: str | None = None
    via_device: "Device | None" = None
    _extra_availability: dict = field(default_factory=dict)

    _unique_id: str | None = None

    __entities: list["Entity"] = field(default_factory=list)

    def __post_init__(self):
        self._unique_id = f"{self.service.SERVICE_NAME}_{clean_string(self.name)}"
        if self.identifiers is None:
            self.identifiers = [self._unique_id]
        else:
            if isinstance(self.identifiers, str):
                self.identifiers = [self.identifiers]
            if self._unique_id not in self.identifiers:
                self.identifiers.append(self._unique_id)

    @property
    def entities(self):
        return self.__entities

    def add_entity(self, entity):
        self.__entities.append(entity)

    def publish_discovery(self, prefix):
        device_payload = {
            k: v
            for k, v in self.__dict__.items()
            if v is not None and not k.startswith("_")
        }

        del device_payload["service"]

        if self.via_device is not None:
            device_payload["via_device"] = self.via_device._unique_id
        availability = [
            {
                "topic": self.service.willtopic,
                "payload_available": "1",
                "payload_not_available": "0",
            }
        ]
        if self._extra_availability:
            availability.append(self._extra_availability)

        payload = {
            "device": device_payload,
            "origin": {"name": f"Bluebattery service"},
            "availability": availability,
            "components": {},
        }

        for entity in self.__entities:
            payload["components"][entity.unique_id] = entity.get_discover_payload()
        topic = f"{prefix}/device/{self._unique_id}/config"
        self.service.publish_json(topic, payload, qos=1, retain=True)


@dataclass
class EntityWithoutStateTopic:
    device: Device
    name: str
    display_name: str | None = None
    device_class: str | None = None
    enabled_by_default: bool | None = None
    entity_category: str | None = None
    expire_after: int | None = None
    force_update: bool | None = None
    icon: str | None = None
    default_entity_id: str | None = None
    qos: int | None = None
    unique_id: str | None = None

    json_attributes_topic_postfix: str | None = None
    json_attributes_template: str | None = None

    def __post_init__(self):
        if self.default_entity_id is None:
            self.default_entity_id = f"{self._component}.{clean_string(self.name)}"
        if self.unique_id is None:
            self.unique_id = (
                f"{self.device._unique_id}__{self.default_entity_id.replace('.', '_')}"
            )

        self.device.add_entity(self)

    def get_discover_payload(self):
        payload = {
            k: v
            for k, v in self.__dict__.items()
            if v is not None and not k.startswith("_") and k != "device"
        }
        payload["platform"] = self._component

        for key, value in list(payload.items()):
            if key.endswith("_postfix"):
                full_key = key[:-8]
                payload[full_key] = f"{self.device.service.data_topic_prefix}{value}"
                del payload[key]

        return payload


@dataclass
class Entity(EntityWithoutStateTopic):
    state_topic_postfix: str = ""
    value_template: str | None = None

    def __post_init__(self):
        if self.state_topic_postfix == "":
            raise ValueError("state_topic_postfix must be set")
        if self.default_entity_id is None:
            self.default_entity_id = (
                f"{self._component}.{self.state_topic_postfix.replace('/', '_')}"
            )
        if self.unique_id is None:
            self.unique_id = (
                f"{self.device._unique_id}__{self.default_entity_id.replace('.', '_')}"
            )
        self.device.add_entity(self)


@dataclass
class Sensor(Entity):
    _component: str = "sensor"
    unit_of_measurement: str | None = None
    state_class: str | None = None
    last_reset_value_template: str | None = None
    suggested_display_precision: int | None = None
    options: list | None = None

    def __post_init__(self):
        super().__post_init__()

        if self.options is not None:
            self.device_class = "enum"



"""
An MQTT output plugin to publish data to an MQTT broker.
"""

import asyncio
import logging
import json

import paho.mqtt.client as mqtt

from bluebattery.frametypes import ALL_FRAME_TYPES
from bluebattery.ha_sensors import Device, Sensor, clean_string


class MQTTOutput:
    SERVICE_NAME = "bluebattery_py"

    @staticmethod
    def add_subparser(parser):
        mqtt_parser = parser.add_parser("mqtt", help="MQTT output")
        mqtt_parser.add_argument(
            "--host",
            default="localhost",
            help="MQTT broker hostname (default: localhost)",
        )
        mqtt_parser.add_argument(
            "--port", default=1883, type=int, help="MQTT broker port (default: 1883)"
        )
        mqtt_parser.add_argument(
            "--username", default=None, help="MQTT broker username (default: None)"
        )
        mqtt_parser.add_argument(
            "--password", default=None, help="MQTT broker password (default: None)"
        )
        mqtt_parser.add_argument(
            "--topic",
            default="service/bluebattery/",
            help="MQTT topic to publish to will be extended by the device UUID. (default: service/bluebattery)",
        )
        mqtt_parser.add_argument(
            "--client-id",
            default="bluebattery",
            help="MQTT client ID (default: bluebattery)",
        )

    def __init__(self, args):
        self.log = logging.getLogger("output.mqtt")
        self.ha_devices = {}
        self.data_topic_prefix = args.topic
        self.willtopic = f"{self.data_topic_prefix}online"
        self.client = mqtt.Client(args.client_id)
        if args.username:
            self.client.username_pw_set(args.username, args.password)
        self.client.connect(args.host, args.port)
        self.client.loop_start()

        # publish to the "online" topic as last will
        self.client.will_set(self.willtopic, "0", retain=True)
        self.client.publish(self.willtopic, "1", retain=True)

    def callback(self, device, data):
        frame, output_id, output_data = data

        if not device.address in self.ha_devices:
            self.log.info(f"Creating Home Assistant device for {device.address}")
            self.create_ha_device(device)

        for key, value in output_data.items():
            topic = f"{self.data_topic_prefix}{device.address}/{output_id}/{key}"
            if type(value) not in (str, int, float):
                value = str(value)
            self.log.debug(f"Publishing {topic} = {value}")
            self.client.publish(topic, value, retain=False)

    def publish_json(self, topic, payload, qos=0, retain=False):
        self.log.debug(f"Publishing JSON to {topic}: {payload}")
        self.client.publish(topic, json.dumps(payload), qos=qos, retain=retain)

    def create_ha_device(self, device):
        ha_device = Device(
            service=self,
            name=f"{device.name} ({device.address})",
            identifiers=[device.address],
            manufacturer="BlueBattery",
            model="Unknown",
            _extra_availability={
                "topic": f"{self.data_topic_prefix}{device.address}/status/connected",
                "payload_available": "1",
                "payload_not_available": "0",
            },
        )
        self.ha_devices[device.address] = ha_device

        # iterate through all frame types and their commands to create sensors

        for frame_type in ALL_FRAME_TYPES:
            for value in frame_type.fields:
                if not value.ha_details:
                    continue
                if value.ha_details.get("platform") != "sensor":
                    raise NotImplementedError(
                        "Only sensor platform is implemented yet."
                    )
                sensor = Sensor(
                    device=ha_device,
                    name=value.ha_details["name"],
                    device_class=value.ha_details.get("device_class"),
                    unit_of_measurement=value.ha_details.get("unit_of_measurement"),
                    suggested_display_precision=value.ha_details.get(
                        "suggested_display_precision"
                    ),
                    state_topic_postfix=f"{device.address}/{frame_type.output_id}/{value.output_id}",
                    default_entity_id=f"sensor.{clean_string(ha_device.name)}_{clean_string(frame_type.output_id)}_{clean_string(value.output_id)}",
                    unique_id=f"{clean_string(ha_device.name)}__sensor_{clean_string(frame_type.output_id)}_{clean_string(value.output_id)}",
                    enabled_by_default=value.ha_details.get("enabled_by_default", True),
                    state_class=value.ha_details.get("state_class", None),
                )

        ha_device.publish_discovery("homeassistant")
